// sample5.cpp  Copyright (c) 2006 Kosei Demura 
// 2006-2-5
// My web site is http://demura.net
// This program uses the Open Dynamics Engine (ODE) by Russell Smith.
// The ODE web site is http://ode.org/

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCappedCylinder dsDrawCappedCylinderD
#endif

static dWorldID      world;
static dSpaceID      space; 
static dGeomID       ground;
static dJointGroupID contactgroup; 
dsFunctions fn;

typedef struct {   
  dBodyID body;    
  dGeomID geom;    
} MyObject;        
MyObject ball,pillar;     

// Create a ball
void createBall() {
  dMass m1;
  dReal x0 = 0.0, y0 = 0.5, z0 = 2.0, radius = 0.1; //[m]
  dReal mass = 1.0; //[kg]

  ball.body = dBodyCreate(world);   // changed
  dMassSetZero(&m1);
  dMassSetSphereTotal(&m1,mass,radius);
  dBodySetMass(ball.body,&m1);             // changed
  dBodySetPosition(ball.body, x0, y0, z0); // changed

  ball.geom = dCreateSphere(space,radius); // add
  dGeomSetBody(ball.geom,ball.body);       // add
}

// Create a pillar
void createPillar() {
  dMass m1;
  dReal x0 = 0.0, y0 = -0.5, z0 = 2.0;
  dReal ax =1.0, ay=0.0, az=0.0, angle = 0.25 * M_PI;
  dReal radius = 0.1, length = 0.5, mass = 1.0;

  pillar.body = dBodyCreate(world);
  dMassSetZero(&m1);
  dMassSetCappedCylinderTotal(&m1,mass,3,radius,length);
  dBodySetMass(pillar.body,&m1);
  dBodySetPosition(pillar.body, x0, y0, z0);
  dMatrix3 R;
  dRFromAxisAndAngle(R,ax,ay,az,angle);
  dBodySetRotation(pillar.body,R);
  pillar.geom = dCreateCCylinder(space,radius,length); 
  dGeomSetBody(pillar.geom,pillar.body);            
}

static void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
  const int N = 10;
  dContact contact[N];

  int isGround = ((ground == o1) || (ground == o2));

  int n =  dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
  if (isGround)  {
    for (int i = 0; i < n; i++) {
      contact[i].surface.mode = dContactBounce|dContactSoftERP|dContactSoftCFM;
      contact[i].surface.bounce = 0.5; // (0.0~1.0)
      contact[i].surface.soft_erp = 1.0; 
      contact[i].surface.soft_cfm = 0.0;
      dJointID c = dJointCreateContact(world,contactgroup,&contact[i]);
      dJointAttach (c,dGeomGetBody(contact[i].geom.g1),
		    dGeomGetBody(contact[i].geom.g2));
    }
  }
}

void destroyBall() {
  dBodyDestroy(ball.body);
  dGeomDestroy(ball.geom);
}

void destroyCylinder() {
  dBodyDestroy(pillar.body);
  dGeomDestroy(pillar.geom);
}

void drawObject(dGeomID gID,float red,float green,float blue) 
{
  switch (dGeomGetClass(gID)) {
  case dSphereClass:
    {
     dsSetColor(red,green,blue);
    dsDrawSphere(dGeomGetPosition(gID),dGeomGetRotation(gID),
		 dGeomSphereGetRadius(gID));
    break;
    }
  case dCCylinderClass:
    {
    dReal r,length;
    dsSetColor(red,green,blue);
    dGeomCCylinderGetParams(gID, &r,&length);
    dsDrawCylinder(dGeomGetPosition(gID),dGeomGetRotation(gID),length,r);
    break;
    }
  }
}

static void simLoop (int pause)
{
  if (!pause) {
    dSpaceCollide(space,0,&nearCallback); 
    dWorldStep(world,0.01);
    dJointGroupEmpty(contactgroup);
  }    
   
  drawObject(ball.geom,1.3,0,0);
  drawObject(pillar.geom,0,0,1.3);

}

void start()
{
  static float xyz[3] = {   3.0, 0.0,1.0};
  static float hpr[3] = {-180.0, 0.0,0.0};
  dsSetViewpoint (xyz,hpr);
}

void restart()
{
  // destroy
  dJointGroupDestroy(contactgroup);
  destroyBall();
  destroyCylinder();

  // create
  contactgroup = dJointGroupCreate(0); 
  createBall();
  createPillar();
}

void command(int cmd)
{
  switch (cmd) {
  case 's':dBodyAddForce(pillar.body,50,  0, 0);break;
  case 'j':dBodyAddForce(ball.body  , 0,-10, 0);break;
  case 'k':dBodyAddForce(ball.body  , 0,  0,50);break;
  case 'l':dBodyAddForce(ball.body  , 0, 10, 0);break;
  case 'r':restart()                           ;break;
  }
}

void  prepDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = &command;
  fn.path_to_textures = "../../drawstuff/textures";
}

int main (int argc, char **argv)
{
  prepDrawStuff();

  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0); 

  dWorldSetGravity(world,0,0,-0.5);

  ground = dCreatePlane(space,0,0,1,0); 

  createBall();      // create a ball
  createPillar();    // create a pillar

  dsSimulationLoop (argc,argv,640,480,&fn);

  dJointGroupDestroy(contactgroup);
  dSpaceDestroy(space);
  dWorldDestroy(world);
  return 0;
}
